// hlavickove subory *******************************************************
#include "../includes.h"
// vlastny .h subor
#include "RB.h"

// inicializacia RB ********************************************************
void CRigidBody::rb_init(vec3 hS)
{
 // rotacna matica RB - quaternion
 set_R(1.0f,0.0f,0.0f,0.0f);

 // sila a krutiaci moment posobiace na RB
 clear_F_T();

 // zrychlenie, uhlove zrychlenie RB
 clear_a_aa();

 // casovy prirastok rychlosti RB
 delta_v.clear();

 // priamociara a uhlova rychlost RB
 v.clear(); w.clear();

 // kineticka energia RB
 energy=2.0f*inactive_energy;
 // energia RB, pri ktorej RB zmeni stav na neaktivny
 set_inactive_energy(0.3f);
 // RB je aktivne
 set_is_active();
 // RB moze byt neaktivne
 set_can_be_inactive();

 // vrcholy RB
 vrts[0].set(-hS.x,-hS.y, hS.z); vrts[1].set( hS.x,-hS.y, hS.z);
 vrts[2].set( hS.x, hS.y, hS.z); vrts[3].set(-hS.x, hS.y, hS.z);
 vrts[4].set(-hS.x,-hS.y,-hS.z); vrts[5].set( hS.x,-hS.y,-hS.z);
 vrts[6].set( hS.x, hS.y,-hS.z); vrts[7].set(-hS.x, hS.y,-hS.z);

 // 12 trojuholnikov, z ktorych je zlozene RB
 tris.resize(12);
 tris_num=tris.length();
 // vpredu
 tris[0] =triangle(vrts[0],vrts[1],vrts[2],vec2(0.5f,0.5f),vec2(0.75f,0.5f), vec2(0.75f,0.25f));
 tris[1] =triangle(vrts[0],vrts[2],vrts[3],vec2(0.5f,0.5f),vec2(0.75f,0.25f),vec2(0.5f,0.25f));
 // hore
 tris[2] =triangle(vrts[3],vrts[2],vrts[6],vec2(0.5f,0.25f),vec2(0.75f,0.25f),vec2(0.75f,0.0f));
 tris[3] =triangle(vrts[3],vrts[6],vrts[7],vec2(0.5f,0.25f),vec2(0.75f,0.0f), vec2(0.5f,0.0f));
 // vzadu
 tris[4] =triangle(vrts[7],vrts[6],vrts[5],vec2(0.0f,0.5f),vec2(0.25f,0.5f), vec2(0.25f,0.25f));
 tris[5] =triangle(vrts[7],vrts[5],vrts[4],vec2(0.0f,0.5f),vec2(0.25f,0.25f),vec2(0.0f,0.25f));
 // dole
 tris[6] =triangle(vrts[4],vrts[5],vrts[1],vec2(0.5f,0.75f),vec2(0.75f,0.75f),vec2(0.75f,0.5f));
 tris[7] =triangle(vrts[4],vrts[1],vrts[0],vec2(0.5f,0.75f),vec2(0.75f,0.5f), vec2(0.5f,0.5f));
 // vlavo
 tris[8] =triangle(vrts[4],vrts[0],vrts[3],vec2(0.25f,0.5f),vec2(0.5f,0.5f), vec2(0.5f,0.25f));
 tris[9] =triangle(vrts[4],vrts[3],vrts[7],vec2(0.25f,0.5f),vec2(0.5f,0.25f),vec2(0.25f,0.25f));
 // vpravo
 tris[10]=triangle(vrts[1],vrts[5],vrts[6],vec2(0.75f,0.5f),vec2(1.0f,0.5f), vec2(1.0f,0.25f));
 tris[11]=triangle(vrts[1],vrts[6],vrts[2],vec2(0.75f,0.5f),vec2(1.0f,0.25f),vec2(0.75f,0.25f));
}

// normalizacia R, vypocet t_mat, I_inv - vzdy po zmene R, alebo Ibs *******
void CRigidBody::update_params(void)
{
 // normalizacia R
 R.normalize();

 // vypocet t_mat - transformacna matica z quaternionu
 // quaternion q = [ s, vx, vy, vz ] = [ r, i, j, k ]
 // t_mat = 1-2*vy^2 - 2*vz^2,  2*vx*vy - 2*s*vz,  2*vx*vz + 2*s*vy
 //          2*vx*vy + 2*s*vz, 1-2*vx^2 - 2*vz^2,  2*vy*vz - 2*s*vx
 //          2*vx*vz - 2*s*vy,  2*vy*vz + 2*s*vx, 1-2*vx^2 - 2*vy^2
 //       = 1-2*j^2 - 2*k^2,   2*i*j - 2*r*k,   2*i*k + 2*r*j
 //           2*i*j + 2*r*k, 1-2*i^2 - 2*k^2,   2*j*k - 2*r*i
 //           2*i*k - 2*r*j,   2*j*k + 2*r*i, 1-2*i^2 - 2*j^2
 t_mat.d[0]= 1.0f-2.0f*R.j*R.j-2.0f*R.k*R.k;
 t_mat.d[1]=      2.0f*R.i*R.j-2.0f*R.r*R.k;
 t_mat.d[2]=      2.0f*R.i*R.k+2.0f*R.r*R.j;
 t_mat.d[3]=s.x;
 t_mat.d[4]=      2.0f*R.i*R.j+2.0f*R.r*R.k;
 t_mat.d[5]= 1.0f-2.0f*R.i*R.i-2.0f*R.k*R.k;
 t_mat.d[6]=      2.0f*R.j*R.k-2.0f*R.r*R.i;
 t_mat.d[7]=s.y;
 t_mat.d[8]=      2.0f*R.i*R.k-2.0f*R.r*R.j;
 t_mat.d[9]=      2.0f*R.j*R.k+2.0f*R.r*R.i;
 t_mat.d[10]=1.0f-2.0f*R.i*R.i-2.0f*R.j*R.j;
 t_mat.d[11]=s.z;

 // vypocet I_inv - inverzny tenzor zotrvacnosti vo world space
 // I_inv = R * Ibs_inv * R_transp
 //               0 1  2  3   0 1 2
 // R * Ibs_inv = 4 5  6  7 * 3 4 5
 //               8 9 10 11   6 7 8
 t[0]=t_mat.d[0]*Ibs_inv.d[0]+t_mat.d[1]*Ibs_inv.d[3]+t_mat.d[2]*Ibs_inv.d[6];
 t[1]=t_mat.d[0]*Ibs_inv.d[1]+t_mat.d[1]*Ibs_inv.d[4]+t_mat.d[2]*Ibs_inv.d[7];
 t[2]=t_mat.d[0]*Ibs_inv.d[2]+t_mat.d[1]*Ibs_inv.d[5]+t_mat.d[2]*Ibs_inv.d[8];
 t[3]=t_mat.d[4]*Ibs_inv.d[0]+t_mat.d[5]*Ibs_inv.d[3]+t_mat.d[6]*Ibs_inv.d[6];
 t[4]=t_mat.d[4]*Ibs_inv.d[1]+t_mat.d[5]*Ibs_inv.d[4]+t_mat.d[6]*Ibs_inv.d[7];
 t[5]=t_mat.d[4]*Ibs_inv.d[2]+t_mat.d[5]*Ibs_inv.d[5]+t_mat.d[6]*Ibs_inv.d[8];
 t[6]=t_mat.d[8]*Ibs_inv.d[0]+t_mat.d[9]*Ibs_inv.d[3]+t_mat.d[10]*Ibs_inv.d[6];
 t[7]=t_mat.d[8]*Ibs_inv.d[1]+t_mat.d[9]*Ibs_inv.d[4]+t_mat.d[10]*Ibs_inv.d[7];
 t[8]=t_mat.d[8]*Ibs_inv.d[2]+t_mat.d[9]*Ibs_inv.d[5]+t_mat.d[10]*Ibs_inv.d[8];
 //                              0 1 2   0 4  8
 // ( R * Ibs_inv ) * R_transp = 3 4 5 * 1 5  9
 //                              6 7 8   2 6 10
 //                                      3 7 11
 I_inv.d[0]=t[0]*t_mat.d[0]+t[1]*t_mat.d[1]+t[2]*t_mat.d[2];
 I_inv.d[1]=t[0]*t_mat.d[4]+t[1]*t_mat.d[5]+t[2]*t_mat.d[6];
 I_inv.d[2]=t[0]*t_mat.d[8]+t[1]*t_mat.d[9]+t[2]*t_mat.d[10];
 I_inv.d[3]=t[3]*t_mat.d[0]+t[4]*t_mat.d[1]+t[5]*t_mat.d[2];
 I_inv.d[4]=t[3]*t_mat.d[4]+t[4]*t_mat.d[5]+t[5]*t_mat.d[6];
 I_inv.d[5]=t[3]*t_mat.d[8]+t[4]*t_mat.d[9]+t[5]*t_mat.d[10];
 I_inv.d[6]=t[6]*t_mat.d[0]+t[7]*t_mat.d[1]+t[8]*t_mat.d[2];
 I_inv.d[7]=t[6]*t_mat.d[4]+t[7]*t_mat.d[5]+t[8]*t_mat.d[6];
 I_inv.d[8]=t[6]*t_mat.d[8]+t[7]*t_mat.d[9]+t[8]*t_mat.d[10];
}

// vypocet casoveho kroku fyzikalnej simulacie *****************************
void CRigidBody::integrate(float DT)
{
 // ak je RB neaktivne, vypocet sa nevykovava,
 // RB sa musi aktivovat posobenim externych sil
 if (!is_active) return;

 // priamociare zrychlenie na zaklade posobiacich sil
 // a += F / M
 a+=F*M_inv;
 // casovy prirastok rychlosti
 // delta_v = a * delta t
 delta_v=a*DT;
 // priamociara rychlost
 v+=delta_v;

 // uhlove zrychlenie na zaklade posobenia krutiaceho momentu
 // aa += I_inv * T
 aa+=I_inv.mul(T);
 // uhlova rychlost
 w+=aa*DT;

 // casovo zavisly utlm
 // ak delta t = 0.0f potom pow(l_damping,DT) = 1.0f => bez utlmu
 // ak delta t = 1.0f potom pow(l_damping,DT) = l_damping => utlm
 v*=pow(l_damping,DT); w*=pow(a_damping,DT);

 // poloha
 s+=v*DT;

 // rotacna matica
 R.add_scaled_vector(w,DT);

 // normalizacia R, vypocet t_mat, I_inv - vzdy po zmene R, alebo Ibs
 update_params();

 // zrusenie posobiacich sil
 clear_F_T();

 // deaktivacia RB v pripade malej kinetickej energie
 if (can_be_inactive) {
  act_energy=v.squaredlen()+w.squaredlen();
  // ratio = 0.5f ^ delta t
  // ak delta t = 0.0f ratio = 1.0f => maximum predoslej energie
  // ak delta t = 1.0f ratio = 0.5f => 1/2 predoslej energie + 1/2 aktualnej energie
  ratio=pow(0.5f,DT);
  energy=ratio*energy+(1.0f-ratio)*act_energy;

  if (energy<inactive_energy) set_is_active(false); // RB bude neaktivne
  else energy=min(energy,2.0f*inactive_energy);
 }
}
